from dataclasses import dataclass
from typing import (
    Optional,
    List,
    Dict,
)

from bdbt.ethereum.abi.abi_type import ABIEvent, ABICall

try:
    from typing import TypedDict
except ImportError:
    from typing_extensions import TypedDict

from bdbt.ethereum.exceptions import ABITypeNotValid


@dataclass
class ABIDataType:
    """
    Follow by: https://docs.soliditylang.org/en/v0.8.11/abi-spec.html#types
    """
    canonical_type: Optional[str] = None


class ABIIntType(ABIDataType):
    def __init__(self, bit_length: int, unsigned: bool):
        if bit_length > 256 or bit_length <= 0 or bit_length % 8 != 0:
            raise ABITypeNotValid(
                'The bit length of the integer type must less than or equal to 256, more than 0 and divisible by 8.')

        super(ABIIntType, self).__init__(canonical_type=f'{"u" if unsigned else ""}int{bit_length}')
        self.bit_length = bit_length
        self.unsigned = unsigned


class ABIStringType(ABIDataType):
    def __init__(self):
        super(ABIStringType, self).__init__(canonical_type='string')


class ABIAddressType(ABIDataType):
    """
    The address type equivalent to uint160 in the document, but we will transform to HexString to show.
    """

    def __init__(self):
        super(ABIAddressType, self).__init__(canonical_type='address')


class ABIBoolType(ABIDataType):
    def __init__(self):
        super(ABIBoolType, self).__init__(canonical_type='bool')


class ABIFixedType(ABIDataType):
    def __init__(self, bit_length: int, scale: int, unsigned: bool):
        if bit_length > 256 or bit_length < 8 or bit_length % 8 != 0:
            raise ABITypeNotValid(
                'The bit length of the fixed type must less than or equal to 256, more than 0 and divisible by 8.')

        if scale > 80 or scale <= 0:
            raise ABITypeNotValid(
                'The scale of the fixed type must less than or equal to 80 and more than 0.')

        super(ABIFixedType, self).__init__(canonical_type=f'{"u" if unsigned else ""}fixed{bit_length}x{scale}')
        self.bit_length = bit_length
        self.scale = scale
        self.unsigned = unsigned


class ABIByteType(ABIDataType):
    def __init__(self):
        super(ABIByteType, self).__init__(canonical_type='byte')


class ABIArrayType(ABIDataType):
    def __init__(self, canonical_type: str, element_type: ABIDataType, length: int):
        super(ABIArrayType, self).__init__(canonical_type=canonical_type)

        self.length = length
        self.element_type = element_type


class ABIBytesType(ABIArrayType):
    def __init__(self, length: int, dynamic: bool):
        if length > 32 or length <= 0:
            raise ABITypeNotValid(
                'The length of bytes must les than or equal to 32 and more than 0.')

        super(ABIBytesType, self).__init__(
            canonical_type=f'bytes{length}',
            element_type=ABIByteType(),
            length=length
        )

        self.dynamic = dynamic


class ABIFunctionType(ABIArrayType):
    def __init__(self):
        super(ABIFunctionType, self).__init__(
            canonical_type='function',
            element_type=ABIByteType(),
            length=24
        )


class ABITupleType(ABIDataType):
    def __init__(self, element_fields: List['ABIField']):
        super().__init__(canonical_type='tuple')

        self.element_fields = element_fields

    def add(self, field: 'ABIField'):
        self.element_fields.append(field)


@dataclass
class ABIField:
    name: str
    ftype: ABIDataType
    metadata: Optional[Dict[str, any]] = None


@dataclass
class ABIEventSchema:
    name: str
    inputs: List[ABIField]
    raw_schema: ABIEvent

    @property
    def is_empty(self):
        return len(self.inputs) == 0


@dataclass
class ABICallSchema:
    name: str
    inputs: List[ABIField]
    outputs: List[ABIField]
    raw_schema: ABICall

    @property
    def is_empty(self):
        return len(self.inputs) + len(self.outputs) == 0


@dataclass
class ABISchema:
    events: List[ABIEventSchema]
    calls: List[ABICallSchema]

    @property
    def nonempty_events(self):
        return [i for i in self.events if not i.is_empty]

    @property
    def nonempty_calls(self):
        return [i for i in self.calls if not i.is_empty]



from bdbt.ethereum.dbt.dbt_code_generator import DbtCodeGenerator
from bdbt.ethereum.dbt.spark.spark_dbt_code_generator import SparkDbtCodeGenerator
from bdbt.global_type import Database


class DbtFactory:

    @staticmethod
    def new_code_generator(database: Database, remote_workspace: str) -> DbtCodeGenerator:
        if database == Database.SPARK:
            return SparkDbtCodeGenerator(remote_workspace)
        else:
            raise ValueError(f'{database} is not be supported now.')



from bdbt.ethereum.abi.abi_data_type import (
    ABIArrayType,
    ABIFunctionType,
    ABIBytesType,
    ABIFixedType,
    ABIBoolType,
    ABIAddressType,
    ABIStringType,
    ABIIntType,
    ABITupleType
)
from bdbt.ethereum.abi.provider.data_type_provider import DataTypeProvider


class HiveObjectInspectorTypeProvider(DataTypeProvider[str]):
    def transform_from_int_type(self, atype: ABIIntType) -> str:
        signed_bit_length = atype.bit_length + atype.unsigned

        if 0 < signed_bit_length <= 32:
            return 'PrimitiveObjectInspectorFactory.writableIntObjectInspector'
        elif 32 < signed_bit_length <= 64:
            return 'PrimitiveObjectInspectorFactory.writableLongObjectInspector'
        elif 64 < signed_bit_length:
            return 'PrimitiveObjectInspectorFactory.writableHiveDecimalObjectInspector'

    def transform_from_string_type(self, atype: ABIStringType) -> str:
        return 'PrimitiveObjectInspectorFactory.writableStringObjectInspector'

    def transform_from_address_type(self, atype: ABIAddressType) -> str:
        return 'PrimitiveObjectInspectorFactory.writableStringObjectInspector'

    def transform_from_bool_type(self, atype: ABIBoolType) -> str:
        return 'PrimitiveObjectInspectorFactory.writableBooleanObjectInspector'

    def transform_from_fixed_type(self, atype: ABIFixedType) -> str:
        return 'PrimitiveObjectInspectorFactory.writableHiveDecimalObjectInspector'

    def transform_from_bytes_type(self, atype: ABIBytesType) -> str:
        return 'PrimitiveObjectInspectorFactory.writableBinaryObjectInspector'

    def transform_from_function_type(self, atype: ABIFunctionType) -> str:
        return 'PrimitiveObjectInspectorFactory.writableBinaryObjectInspector'

    def transform_from_array_type(self, atype: ABIArrayType) -> str:
        return f'ObjectInspectorFactory.getStandardListObjectInspector({self.transform(atype.element_type)})'

    def transform_from_tuple_type(self, atype: ABITupleType) -> str:
        field_names = ','.join([f'"{i.name}"' for i in atype.element_fields])
        field_ois = ',\n'.join([self.transform(i.ftype) for i in atype.element_fields])

        return f"""ObjectInspectorFactory.getStandardStructObjectInspector(
            ImmutableList.of({field_names}),
            ImmutableList.of({field_ois})
        )
        """



import math
from typing import Union

from pyspark.sql.types import (
    DataType,
    IntegerType,
    LongType,
    DecimalType,
    StringType,
    BooleanType,
    BinaryType,
    ArrayType,
    StructType,
    StructField
)

from bdbt.ethereum.abi.abi_data_type import (
    ABIFunctionType,
    ABIBytesType,
    ABIFixedType,
    ABIBoolType,
    ABIAddressType,
    ABIStringType,
    ABIIntType,
    ABITupleType,
    ABIArrayType
)
from bdbt.ethereum.abi.provider.data_type_provider import DataTypeProvider


class SparkDataTypeProvider(DataTypeProvider[DataType]):

    def transform_from_array_type(self, atype: ABIArrayType, deep: int = 0) -> ArrayType:
        return ArrayType(elementType=self.transform(atype.element_type))

    def transform_from_tuple_type(self, atype: ABITupleType) -> StructType:
        return StructType(fields=[StructField(
            name=i.name,
            dataType=self.transform(i.ftype),
            metadata=i.metadata
        ) for i in atype.element_fields])

    def transform_from_int_type(self, atype: ABIIntType) -> Union[IntegerType, LongType, DecimalType]:
        signed_bit_length = atype.bit_length + atype.unsigned

        # Ignore ByteType(8bit) and ShortType(16bit) for match the type in the headlong.
        if 0 < signed_bit_length <= 32:
            return IntegerType()
        elif 32 < signed_bit_length <= 64:
            return LongType()
        elif 64 < signed_bit_length:
            return DecimalType(38, 0)

    def transform_from_string_type(self, atype: ABIStringType) -> StringType:
        return StringType()

    def transform_from_address_type(self, atype: ABIAddressType) -> StringType:
        return StringType()

    def transform_from_bool_type(self, atype: ABIBoolType) -> BooleanType:
        return BooleanType()

    def transform_from_fixed_type(self, atype: ABIFixedType) -> DecimalType:
        signed_bit_length = atype.bit_length + atype.unsigned
        # precision = log10(2^bit_length) + 1
        # The precision can be up to 38, scale can also be up to 38 (less or equal to precision) in Spark
        # doc: https://spark.apache.org/docs/latest/api/java/org/apache/spark/sql/types/DecimalType.html
        precision = min(38, math.floor(signed_bit_length / math.log2(10)) + 1)
        return DecimalType(precision=min(38, math.floor(signed_bit_length / math.log2(10)) + 1),
                           scale=min(precision, atype.scale))

    def transform_from_bytes_type(self, atype: ABIBytesType) -> BinaryType:
        return BinaryType()

    def transform_from_function_type(self, atype: ABIFunctionType) -> BinaryType:
        return BinaryType()



import re
from copy import deepcopy
from typing import List, Dict

from bdbt.ethereum.abi.abi_data_type import (
    ABIDataType,
    ABIIntType,
    ABIAddressType,
    ABIFixedType,
    ABIBoolType,
    ABIBytesType,
    ABIFunctionType,
    ABIStringType,
    ABIField,
    ABIArrayType,
    ABITupleType,
    ABIEventSchema,
    ABICallSchema,
    ABISchema
)
from bdbt.ethereum.abi.abi_type import (
    ABIEventElement,
    ABI,
    ABIEvent,
    ABICall,
    ABICallElement
)
from bdbt.ethereum.abi.utils import filter_by_type_and_name, filter_by_type
from bdbt.ethereum.exceptions import TargetItemNotFound


class ABITransformer:
    abi_type_mapping: Dict[str, ABIDataType] = {}

    for i in range(8, 257, 8):
        abi_type_mapping[f'int{i}'] = ABIIntType(bit_length=i, unsigned=False)
        abi_type_mapping[f'uint{i}'] = ABIIntType(bit_length=i, unsigned=True)

        for j in range(1, 81):
            abi_type_mapping[f'fixed{i}x{j}'] = ABIFixedType(bit_length=i, scale=j, unsigned=False)
            abi_type_mapping[f'ufixed{i}x{j}'] = ABIFixedType(bit_length=i, scale=j, unsigned=True)

    for i in range(1, 33):
        abi_type_mapping[f'bytes{i}'] = ABIBytesType(length=i, dynamic=False)

    abi_type_mapping['int'] = abi_type_mapping['int256']
    abi_type_mapping['uint'] = abi_type_mapping['uint256']
    abi_type_mapping['fixed'] = abi_type_mapping['fixed128x18']
    abi_type_mapping['ufixed'] = abi_type_mapping['ufixed128x18']
    abi_type_mapping['address'] = ABIAddressType()
    abi_type_mapping['bool'] = ABIBoolType()
    abi_type_mapping['bytes'] = ABIBytesType(length=32, dynamic=True)
    abi_type_mapping['function'] = ABIFunctionType()
    abi_type_mapping['string'] = ABIStringType()

    def _transform_event_element(self, event_element: ABIEventElement) -> ABIField:
        name = event_element.get('name')
        atype_str = event_element.get('type')
        indexed = event_element.get('indexed')

        arr_reg = re.search(r'\[[\d]*\]$', atype_str)

        if arr_reg:
            element_type_str = atype_str[:arr_reg.start()]
            element_type = self.abi_type_mapping[element_type_str]
            arr_str = atype_str[arr_reg.start(): arr_reg.end()]
            arr_length = -1 if arr_str == '[]' else int(arr_str[1:-1])
            return ABIField(
                name=name,
                ftype=ABIArrayType(
                    element_type=element_type,
                    length=arr_length,
                    canonical_type=f'{element_type.canonical_type}{arr_str}'
                ),
                metadata={'indexed': indexed}
            )
        else:
            return ABIField(
                name=name,
                ftype=self.abi_type_mapping[atype_str],
                metadata={'indexed': indexed}
            )

    def _transform_call_element(self, function_element: ABICallElement) -> ABIField:
        name = function_element.get('name')
        atype_str = function_element.get('type')
        components = function_element.get('components')

        arr_reg = re.search(r'\[[\d]*\]$', atype_str)

        if components is not None:
            return ABIField(
                name=name,
                ftype=ABITupleType([self._transform_call_element(i) for i in components])
            )
        elif arr_reg:
            element_type_str = atype_str[:arr_reg.start()]
            element_type = self.abi_type_mapping[element_type_str]
            arr_str = atype_str[arr_reg.start(): arr_reg.end()]
            arr_length = -1 if arr_str == '[]' else int(arr_str[1:-1])
            return ABIField(
                name=name,
                ftype=ABIArrayType(
                    element_type=element_type,
                    length=arr_length,
                    canonical_type=f'{element_type.canonical_type}{arr_str}'
                ),
            )
        else:
            return ABIField(
                name=name,
                ftype=self.abi_type_mapping[atype_str],
            )

    def transform_abi_event(self, abi: ABI, event_name: str) -> ABIEventSchema:
        candidate_events: List[ABIEvent] = filter_by_type_and_name(name=event_name, type_str='event', contract_abi=abi)
        if len(candidate_events) != 1:
            raise TargetItemNotFound(f"{event_name} event can not be found in ABI")

        event = candidate_events[0]
        return ABIEventSchema(
            name=event.get('name'),
            inputs=self._revise_fields([self._transform_event_element(i) for i in event.get('inputs', [])]),
            raw_schema=event
        )

    def transform_abi_call(self, abi: ABI, call_name: str) -> ABICallSchema:
        candidate_calls: List[ABICall] = filter_by_type_and_name(name=call_name, type_str='function', contract_abi=abi)
        if len(candidate_calls) != 1:
            raise TargetItemNotFound(f"{call_name} call can not be found in ABI")

        call = candidate_calls[0]
        return ABICallSchema(
            name=call.get('name'),
            inputs=self._revise_fields([self._transform_call_element(i) for i in call.get('inputs', [])]),
            outputs=self._revise_fields([self._transform_call_element(i) for i in call.get('outputs', [])], 'output'),
            raw_schema=call
        )

    def transform_abi(self, abi: ABI) -> ABISchema:
        event_abi_list = filter_by_type(type_str='event', contract_abi=abi)
        call_abi_list = filter_by_type(type_str='function', contract_abi=abi)

        events = [self.transform_abi_event(abi, i['name']) for i in event_abi_list]
        calls = [self.transform_abi_call(abi, i['name']) for i in call_abi_list]
        return ABISchema(events=events, calls=calls)

    @staticmethod
    def _revise_fields(fields: List[ABIField], prefix: str = '') -> List[ABIField]:
        size = len(fields)

        if size == 0:
            return fields

        revised_fields = []
        for i in range(0, size):
            new_field = deepcopy(fields[i])
            if prefix == '':
                new_field.name = f'_{i}' if new_field.name == '' else new_field.name
            else:
                new_field.name = f'{prefix}_{i}' if new_field.name == '' else f'{prefix}_{new_field.name}'
            revised_fields.append(new_field)

        return revised_fields



from enum import Enum
from typing import Sequence, List

try:
    from typing import TypedDict
except ImportError:
    from typing_extensions import TypedDict


class DbtColumn(TypedDict, total=False):
    name: str


class DbtTable(TypedDict, total=False):
    name: str
    columns: Sequence[DbtColumn]


class DbtModelSchema(TypedDict, total=False):
    models: List[DbtTable]


class Database(Enum):
    SPARK = 'spark'
    BIG_QUERY = 'big_query'
    REDSHIFT = 'redshift'
    SNOWFLAKE = 'snowflake'
    POSTGRES = 'postgres'



from typing import Generic, TypeVar

from bdbt.ethereum.abi.abi_data_type import (
    ABIIntType,
    ABIStringType,
    ABIAddressType,
    ABIBoolType,
    ABIFixedType,
    ABIBytesType,
    ABIFunctionType,
    ABIArrayType,
    ABITupleType,
    ABIDataType,
)

T = TypeVar('T')


class DataTypeProvider(Generic[T]):

    def transform(self, atype: ABIDataType) -> T:
        if isinstance(atype, ABIIntType):
            return self.transform_from_int_type(atype)
        elif isinstance(atype, ABIStringType):
            return self.transform_from_string_type(atype)
        elif isinstance(atype, ABIAddressType):
            return self.transform_from_address_type(atype)
        elif isinstance(atype, ABIBoolType):
            return self.transform_from_bool_type(atype)
        elif isinstance(atype, ABIFixedType):
            return self.transform_from_fixed_type(atype)
        elif isinstance(atype, ABIBytesType):
            return self.transform_from_bytes_type(atype)
        elif isinstance(atype, ABIFunctionType):
            return self.transform_from_function_type(atype)
        elif isinstance(atype, ABIArrayType):
            return self.transform_from_array_type(atype)
        elif isinstance(atype, ABITupleType):
            return self.transform_from_tuple_type(atype)

    def transform_from_int_type(self, atype: ABIIntType) -> T:
        raise NotImplementedError()

    def transform_from_string_type(self, atype: ABIStringType) -> T:
        raise NotImplementedError()

    def transform_from_address_type(self, atype: ABIAddressType) -> T:
        raise NotImplementedError()

    def transform_from_bool_type(self, atype: ABIBoolType) -> T:
        raise NotImplementedError()

    def transform_from_fixed_type(self, atype: ABIFixedType) -> T:
        raise NotImplementedError()

    def transform_from_bytes_type(self, atype: ABIBytesType) -> T:
        raise NotImplementedError()

    def transform_from_function_type(self, atype: ABIFunctionType) -> T:
        raise NotImplementedError()

    def transform_from_array_type(self, atype: ABIArrayType) -> T:
        raise NotImplementedError()

    def transform_from_tuple_type(self, atype: ABITupleType) -> T:
        raise NotImplementedError()



from pathlib import Path

import click

from bdbt.ethereum.dbt.dbt_generator import DbtGenerator
from bdbt.global_type import Database


@click.command(context_settings=dict(help_option_names=['-h', '--help']))
@click.option('-d', '--dbt-dir', default=Path.cwd(), show_default=True, type=str,
              help='The absolute path for the dbt project.')
@click.option('-r', '--remote-dir-url', default='s3a://blockchain-dbt/dist/jars', show_default=True, type=str,
              help='The absolute path for remote workspace that will store external dependencies.')
@click.option('-d', '--database', default=Database.SPARK.value, show_default=True, type=str,
              help='The database to work for, like: spark, big_query, snowflake...')
def ethereum_codegen(
        dbt_dir: str = Path.cwd(),
        remote_dir_url: str = 's3a://ifcrypto/blockchain-dbt/jars',
        database: str = Database.SPARK.value,
):
    database_obj = Database(database)
    generator = DbtGenerator(database=database_obj, remote_dir_url=remote_dir_url, dbt_dir=dbt_dir)
    generator.gen_all()



import json
import os.path
import pathlib
import subprocess
import xml.etree.ElementTree as ET

from eth_utils import event_abi_to_log_topic, encode_hex, function_abi_to_4byte_selector

from bdbt.content import Contract
from bdbt.ethereum.abi.abi_data_type import ABIEventSchema, ABICallSchema
from bdbt.ethereum.abi.provider.hive_object_inspector_type_provider import HiveObjectInspectorTypeProvider
from bdbt.ethereum.dbt.dbt_code_generator import DbtCodeGenerator

event_clazz_template = """package io.iftech.sparkudf.hive;

import java.util.List;
import org.apache.hadoop.hive.serde2.objectinspector.ObjectInspector;
import org.apache.hadoop.hive.serde2.objectinspector.ObjectInspectorFactory;
import org.apache.hadoop.hive.serde2.objectinspector.primitive.PrimitiveObjectInspectorFactory;
import org.sparkproject.guava.collect.ImmutableList;

public class {{CLASS_NAME}} extends DecodeContractEventHiveUDF {

    @Override
    public List<String> getInputDataFieldsName() {
        return ImmutableList.of({{FIELD_NAMES}});
    }

    @Override
    public List<ObjectInspector> getInputDataFieldsOIs() {
        return ImmutableList.of({{OBJECT_INSPECTORS}});
    }
}
"""

call_clazz_template = """package io.iftech.sparkudf.hive;

import java.util.List;
import org.apache.hadoop.hive.serde2.objectinspector.ObjectInspector;
import org.apache.hadoop.hive.serde2.objectinspector.ObjectInspectorFactory;
import org.apache.hadoop.hive.serde2.objectinspector.primitive.PrimitiveObjectInspectorFactory;
import org.sparkproject.guava.collect.ImmutableList;

public class {{CLASS_NAME}} extends DecodeContractFunctionHiveUDF {

    @Override
    public List<String> getInputDataFieldsName() {
        return ImmutableList.of({{INPUT_FIELD_NAMES}});
    }

    @Override
    public List<ObjectInspector> getInputDataFieldsOIs() {
        return ImmutableList.of({{INPUT_OBJECT_INSPECTORS}});
    }

    @Override
    public List<String> getOutputDataFieldsName() {
        return ImmutableList.of({{OUTPUT_FIELD_NAMES}});
    }

    @Override
    public List<ObjectInspector> getOutputDataFieldsOIs() {
        return ImmutableList.of({{OUTPUT_OBJECT_INSPECTORS}});
    }
}
"""

empty_event_dbt_model_sql_template = """{{
    config(
        {{MODEL_MATERIALIZED_CONFIG}},
        file_format='parquet',
        alias='{{MODEL_ALIAS}}'
    )
}}

select /*+ REPARTITION({{MODEL_REPARTITION_COUNT}}) */
    block_number as evt_block_number,
    block_timestamp as evt_block_time,
    log_index as evt_index,
    transaction_hash as evt_tx_hash,
    address as contract_address,
    dt
from {{ ref('stg_logs') }}
where address = lower("{{CONTRACT_ADDRESS}}")
and address_hash = abs(hash(lower("{{CONTRACT_ADDRESS}}"))) % 10
and selector = "{{EVENT_SELECTOR}}"
and selector_hash = abs(hash("{{EVENT_SELECTOR}}")) % 10

{% if is_incremental() %}
  and dt = '{{ var("dt") }}'
{% endif %}
"""

event_dbt_model_sql_template = """{{
    config(
        {{MODEL_MATERIALIZED_CONFIG}},
        file_format='parquet',
        alias='{{MODEL_ALIAS}}',
        pre_hook={
            'sql': 'create or replace function {{UDF_NAME}} as "io.iftech.sparkudf.hive.{{CLASS_NAME}}" using jar "{{UDF_JAR_PATH}}";'
        }
    )
}}

with base as (
    select
        block_number as evt_block_number,
        block_timestamp as evt_block_time,
        log_index as evt_index,
        transaction_hash as evt_tx_hash,
        address as contract_address,
        dt,
        {{UDF_NAME}}(unhex_data, topics_arr, '{{EVENT_ABI}}', '{{EVENT_NAME}}') as data
    from {{ ref('stg_logs') }}
    where address = lower("{{CONTRACT_ADDRESS}}")
    and address_hash = abs(hash(lower("{{CONTRACT_ADDRESS}}"))) % 10
    and selector = "{{EVENT_SELECTOR}}"
    and selector_hash = abs(hash("{{EVENT_SELECTOR}}")) % 10

    {% if is_incremental() %}
      and dt = '{{ var("dt") }}'
    {% endif %}
),

final as (
    select
        evt_block_number,
        evt_block_time,
        evt_index,
        evt_tx_hash,
        contract_address,
        dt,
        data.input.*
    from base
)

select /*+ REPARTITION({{MODEL_REPARTITION_COUNT}}) */ *
from final
"""

empty_call_dbt_model_sql_template = """{{
    config(
        {{MODEL_MATERIALIZED_CONFIG}},
        file_format='parquet',
        alias='{{MODEL_ALIAS}}'
    )
}}

select /*+ REPARTITION({{MODEL_REPARTITION_COUNT}}) */
    status==1 as call_success,
    block_number as call_block_number,
    block_timestamp as call_block_time,
    trace_address as call_trace_address,
    transaction_hash as call_tx_hash,
    to_address as contract_address,
    dt
from {{ ref('stg_traces') }}
where to_address = lower("{{CONTRACT_ADDRESS}}")
and address_hash = abs(hash(lower("{{CONTRACT_ADDRESS}}"))) % 10
and selector = "{{CALL_SELECTOR}}"
and selector_hash = abs(hash("{{CALL_SELECTOR}}")) % 10

{% if is_incremental() %}
  and dt = '{{ var("dt") }}'
{% endif %}
"""

call_dbt_model_sql_template = """{{
    config(
        {{MODEL_MATERIALIZED_CONFIG}},
        file_format='parquet',
        alias='{{MODEL_ALIAS}}',
        pre_hook={
            'sql': 'create or replace function {{UDF_NAME}} as "io.iftech.sparkudf.hive.{{CLASS_NAME}}" using jar "{{UDF_JAR_PATH}}";'
        }
    )
}}

with base as (
    select
        status==1 as call_success,
        block_number as call_block_number,
        block_timestamp as call_block_time,
        trace_address as call_trace_address,
        transaction_hash as call_tx_hash,
        to_address as contract_address,
        dt,
        {{UDF_NAME}}(unhex_input, unhex_output, '{{CALL_ABI}}', '{{CALL_NAME}}') as data
    from {{ ref('stg_traces') }}
    where to_address = lower("{{CONTRACT_ADDRESS}}")
    and address_hash = abs(hash(lower("{{CONTRACT_ADDRESS}}"))) % 10
    and selector = "{{CALL_SELECTOR}}"
    and selector_hash = abs(hash("{{CALL_SELECTOR}}")) % 10

    {% if is_incremental() %}
      and dt = '{{ var("dt") }}'
    {% endif %}
),

final as (
    select
        call_success,
        call_block_number,
        call_block_time,
        call_trace_address,
        call_tx_hash,
        contract_address,
        dt,
        data.input.*,
        data.output.*
    from base
)

select /*+ REPARTITION({{MODEL_REPARTITION_COUNT}}) */ *
from final
"""

table_model_config = "materialized='table'"
increment_model_config = "materialized='incremental', incremental_strategy='insert_overwrite', partition_by=['dt']"


class SparkDbtCodeGenerator(DbtCodeGenerator):
    hive_provider = HiveObjectInspectorTypeProvider()

    def __init__(self, remote_workspace: str):
        super(SparkDbtCodeGenerator, self).__init__(True)
        self.remote_workspace = remote_workspace

    def gen_event_dbt_model(
            self,
            project_path: str,
            contract: Contract,
            version: str,
            event: ABIEventSchema
    ) -> None:
        contract_name = contract['name']
        contract_address = contract['address']
        contract_materialize = contract['materialize']

        project_name = pathlib.Path(project_path).name
        filepath = os.path.join(project_path, self.evt_model_name(contract_name, event, project_name) + '.sql')
        event_selector = encode_hex(event_abi_to_log_topic(event.raw_schema))

        if event.is_empty:
            content = event_dbt_model_sql_template \
                .replace('{{CONTRACT_ADDRESS}}', contract_address) \
                .replace('{{EVENT_SELECTOR}}', event_selector) \
                .replace('{{MODEL_ALIAS}}', self.evt_model_name(contract_name, event).lower()) \
                .replace('{{MODEL_MATERIALIZED_CONFIG}}', self._materialized_config(contract_materialize)) \
                .replace('{{MODEL_REPARTITION_COUNT}}', self._repartition_count(contract_materialize))
        else:
            clazz_name = self._event_udf_class_name(project_name, contract_name, event)
            content = event_dbt_model_sql_template \
                .replace('{{UDF_NAME}}', clazz_name.lower()) \
                .replace('{{CLASS_NAME}}', clazz_name) \
                .replace('{{UDF_JAR_PATH}}', os.path.join(self.remote_workspace, self._jar_name(version))) \
                .replace('{{EVENT_ABI}}', json.dumps(event.raw_schema)) \
                .replace('{{EVENT_NAME}}', event.name) \
                .replace('{{CONTRACT_ADDRESS}}', contract_address) \
                .replace('{{EVENT_SELECTOR}}', event_selector) \
                .replace('{{MODEL_ALIAS}}', self.evt_model_name(contract_name, event).lower()) \
                .replace('{{MODEL_MATERIALIZED_CONFIG}}', self._materialized_config(contract_materialize)) \
                .replace('{{MODEL_REPARTITION_COUNT}}', self._repartition_count(contract_materialize))

        self.create_file_and_write(filepath, content)

    def gen_call_dbt_model(
            self,
            project_path: str,
            contract: Contract,
            version: str,
            call: ABICallSchema
    ) -> None:
        contract_name = contract['name']
        contract_address = contract['address']
        contract_materialize = contract['materialize']

        project_name = pathlib.Path(project_path).name
        filepath = os.path.join(project_path, self.call_model_name(contract_name, call, project_name) + '.sql')
        call_selector = encode_hex(function_abi_to_4byte_selector(call.raw_schema))

        if call.is_empty:
            content = empty_call_dbt_model_sql_template \
                .replace('{{CONTRACT_ADDRESS}}', contract_address) \
                .replace('{{CALL_SELECTOR}}', call_selector) \
                .replace('{{MODEL_ALIAS}}', self.call_model_name(contract_name, call).lower()) \
                .replace('{{MODEL_MATERIALIZED_CONFIG}}', self._materialized_config(contract_materialize)) \
                .replace('{{MODEL_REPARTITION_COUNT}}', self._repartition_count(contract_materialize))
        else:
            clazz_name = self._call_udf_class_name(project_name, contract_name, call)
            content = call_dbt_model_sql_template \
                .replace('{{UDF_NAME}}', clazz_name.lower()) \
                .replace('{{CLASS_NAME}}', clazz_name) \
                .replace('{{UDF_JAR_PATH}}', os.path.join(self.remote_workspace, self._jar_name(version))) \
                .replace('{{CALL_ABI}}', json.dumps(call.raw_schema)) \
                .replace('{{CALL_NAME}}', call.name) \
                .replace('{{CONTRACT_ADDRESS}}', contract_address) \
                .replace('{{CALL_SELECTOR}}', call_selector[0:10]) \
                .replace('{{MODEL_ALIAS}}', self.call_model_name(contract_name, call).lower()) \
                .replace('{{MODEL_MATERIALIZED_CONFIG}}', self._materialized_config(contract_materialize)) \
                .replace('{{MODEL_REPARTITION_COUNT}}', self._repartition_count(contract_materialize))

        self.create_file_and_write(filepath, content)

    def gen_event_udf(
            self, udf_workspace: str, project_name: str, contract_name: str, event: ABIEventSchema
    ) -> None:
        clazz_name = self._event_udf_class_name(project_name, contract_name, event)
        filepath = os.path.join(udf_workspace, clazz_name + '.java')

        field_names = ','.join([f'"{i.name}"' for i in event.inputs])
        # TODO: java file indent style
        field_ois = ',\n'.join([self.hive_provider.transform(i.ftype) for i in event.inputs])
        content = event_clazz_template \
            .replace('{{CLASS_NAME}}', clazz_name) \
            .replace('{{FIELD_NAMES}}', field_names) \
            .replace('{{OBJECT_INSPECTORS}}', field_ois)

        self.create_file_and_write(filepath, content)

    def gen_call_udf(
            self, udf_workspace: str, project_name: str, contract_name: str, call: ABICallSchema
    ) -> None:
        clazz_name = self._call_udf_class_name(project_name, contract_name, call)
        filepath = os.path.join(udf_workspace, clazz_name + '.java')

        input_field_names = ','.join([f'"{i.name}"' for i in call.inputs])
        input_field_ois = ',\n'.join([self.hive_provider.transform(i.ftype) for i in call.inputs])

        output_field_names = ','.join([f'"{i.name}"' for i in call.outputs])
        output_field_ois = ',\n'.join([self.hive_provider.transform(i.ftype) for i in call.outputs])
        content = call_clazz_template \
            .replace('{{CLASS_NAME}}', clazz_name) \
            .replace('{{INPUT_FIELD_NAMES}}', input_field_names) \
            .replace('{{INPUT_OBJECT_INSPECTORS}}', input_field_ois) \
            .replace('{{OUTPUT_FIELD_NAMES}}', output_field_names) \
            .replace('{{OUTPUT_OBJECT_INSPECTORS}}', output_field_ois)

        self.create_file_and_write(filepath, content)

    def build_udf(
            self, dbt_dir: str, version: str
    ) -> None:
        java_project_path = os.path.join(dbt_dir, 'java')

        root = ET.parse(os.path.join(java_project_path, 'pom.xml')).getroot()
        blockchain_spark_version = root.find('{http://maven.apache.org/POM/4.0.0}version').text

        # The version of the jar package should be the same with the version of the dbt project.
        jar_path = os.path.join(dbt_dir, self._jar_name(version))

        self._execute_command(
            command=f"""
            mvn clean package -DskipTests \
            && mv target/blockchain-spark-{blockchain_spark_version}-jar-with-dependencies.jar {jar_path} \
            && cd {dbt_dir} \
            && rm -rf java
            """,
            dir=java_project_path
        )

    def prepare_udf_workspace(self, dbt_dir: str) -> str:
        # clone blockchain-spark project and move java dir to the root
        # TODO: release blockchain-spark project to maven central
        self._execute_command(
            command=f"""
            git clone https://github.com/datawaves-xyz/blockchain-spark.git \
            && mv blockchain-spark/java java \
            && rm -rf blockchain-spark
            """,
            dir=dbt_dir
        )
        return os.path.join(dbt_dir, 'java/src/main/java/io/iftech/sparkudf/hive')

    def _execute_command(self, command: str, dir: str):
        sp = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            cwd=dir,
            shell=True,
            close_fds=True)
        self.logger.info("Output:")
        for line in iter(sp.stdout.readline, b''):
            line = line.decode('utf-8').rstrip()
            self.logger.info(line)
        sp.wait()
        self.logger.info(
            "Command exited with return code %s",
            sp.returncode
        )

        if sp.returncode != 0:
            err_msg = '\n'.join([line.decode('utf-8').rstrip() for line in iter(sp.stderr.readline, b'')])
            raise ChildProcessError(f'execute {command} getting error response: {err_msg}')

    @staticmethod
    def _event_udf_class_name(
            project_name: str, contract_name: str, event: ABIEventSchema
    ) -> str:
        return project_name[0].upper() + project_name[1:] \
               + '_' + contract_name + '_' + event.name + '_' + 'EventDecodeUDF'

    @staticmethod
    def _call_udf_class_name(
            project_name: str, contract_name: str, call: ABICallSchema
    ) -> str:
        return project_name[0].upper() + project_name[1:] \
               + '_' + contract_name + '_' + call.name + '_' + 'CallDecodeUDF'

    @staticmethod
    def _jar_name(version: str) -> str:
        return f'blockchain-dbt-udf-{version}.jar'

    @staticmethod
    def _materialized_config(materialize: str) -> str:
        if materialize == 'table':
            return table_model_config
        elif materialize == 'increment':
            return increment_model_config
        else:
            raise ValueError(f'{materialize} isnt a supported materialized model.')

    @staticmethod
    def _repartition_count(materialize: str) -> str:
        # the best partition size in the Spark is 100-200MB,
        # combining the number of files and the file size,
        # 50 is more suitable for the current data size (the largest table is probably 10G)
        if materialize == 'table':
            return '50'
        elif materialize == 'increment':
            return 'dt'
        else:
            raise ValueError(f'{materialize} isnt a supported materialized model.')



import functools
import glob
import json
import logging
import os
import shutil
from os import listdir
from typing import Dict, cast, List

import pyaml
import ruamel.yaml
import yaml

from bdbt.content import Contract
from bdbt.ethereum.abi.abi_data_type import ABISchema
from bdbt.ethereum.abi.abi_transformer import ABITransformer
from bdbt.ethereum.dbt.dbt_code_generator import DbtCodeGenerator as CG
from bdbt.ethereum.dbt.dbt_factory import DbtFactory
from bdbt.ethereum.dbt.dbt_schema_generator import DbtSchemaGenerator
from bdbt.global_type import Database, DbtTable, DbtColumn, DbtModelSchema

evt_base_column = [
    'evt_block_number',
    'evt_block_time',
    'evt_index',
    'evt_tx_hash',
    'contract_address',
    'dt'
]

call_base_column = [
    'call_success',
    'call_block_number',
    'call_block_time',
    'call_trace_address',
    'call_tx_hash',
    'contract_address',
    'dt'
]


class DbtGenerator:
    def __init__(self, database: Database, dbt_dir: str, remote_dir_url: str):
        self._dbt_dir = dbt_dir
        self._remote_dir_url = remote_dir_url
        self._codegen = DbtFactory.new_code_generator(database, remote_dir_url)
        self._schemagen = DbtSchemaGenerator()
        self._transformer = ABITransformer()
        self._logger = logging.getLogger(self.__class__.__name__)

    def gen_all(self):
        # Remove the old codegen folder and recreate it
        if os.path.exists(self.codegen_dir):
            shutil.rmtree(self.codegen_dir)
        os.mkdir(self.codegen_dir)
        self._logger.info('recreate codegen folder.')

        self._gen_models_and_schema()
        self._gen_udf()
        self._replenish_project_yml()

    def _gen_models_and_schema(self):
        models_count_map: Dict[str, int] = {}

        for project, contracts in self.contracts_map.items():
            # Generate models
            for contract in contracts:
                self._codegen.gen_models_for_project(
                    workspace=self.codegen_dir,
                    project_name=project,
                    contract=contract,
                    version=self.version,
                    abi=self._transformer.transform_abi(contract['abi'])
                )

            # Generate schema
            models: List[DbtTable] = []
            for contract in contracts:
                abi = self._transformer.transform_abi(contract['abi'])
                for event in abi.events:
                    columns = [DbtColumn(name=i.name) for i in event.inputs]
                    columns.extend(DbtColumn(name=i) for i in evt_base_column)
                    models.append(DbtTable(name=CG.evt_model_name(contract['name'], event, project),
                                           columns=columns))

                for call in abi.calls:
                    columns = [DbtColumn(name=i.name) for i in call.inputs]
                    columns.extend([DbtColumn(name=i.name) for i in call.outputs])
                    columns.extend(DbtColumn(name=i) for i in call_base_column)
                    models.append(DbtTable(name=CG.call_model_name(contract['name'], call, project),
                                           columns=columns))

            schema = DbtModelSchema(models=models)
            schema_path = os.path.join(self.codegen_dir, project, 'schema.yml')
            with open(schema_path, 'w') as f:
                # https://docs.getdbt.com/faqs/why-version-2
                f.write('version: 2\n')
                f.write(pyaml.dump(schema, sort_dicts=False))

            models_count_map[project] = len(models)

        self._logger.info('generate all models and schemas: ')
        for project, count in models_count_map.items():
            self._logger.info(f'  {project} has {count} models')

    def _gen_udf(self):
        if not self._codegen.need_udf:
            return

        abi_map: Dict[str, Dict[str, ABISchema]] = {}
        for project, contracts in self.contracts_map.items():
            abi_map[project] = {}
            for contract in contracts:
                abi_map[project][contract['name']] = self._transformer.transform_abi(contract['abi'])

        self._codegen.gen_udf_for_dbt(self._dbt_dir, abi_map, self.version)
        self._logger.info('generate a UDF dependency.')

    def _replenish_project_yml(self):
        projects_dict = {
            project: {
                '+schema': project,
                '+tags': ['chain_ethereum', 'level_parse', f'proj_{project}']
            }
            for project in self.contracts_map.keys()
        }

        project_conf, ind, bsi = ruamel.yaml.util.load_yaml_guess_indent(open(self.dbt_project_yml))
        project_conf = json.loads(json.dumps(project_conf))

        if 'models' not in project_conf.keys():
            raise ValueError('models should be in the dbt_project.yml')
        else:
            model_conf: Dict[str, any] = project_conf['models']

        if len(model_conf.keys()) != 1:
            raise ValueError('only one database should be defined in the models')
        else:
            database_conf: Dict[str, any] = model_conf[list(model_conf.keys())[0]]

        database_conf['codegen'] = projects_dict

        y = ruamel.yaml.YAML()
        y.indent(mapping=ind, sequence=ind, offset=bsi)
        with open(self.dbt_project_yml, 'w') as nf:
            y.dump(project_conf, nf)

        self._logger.info('replenish dbt_project.yml with all project configs.')

    @property
    def contracts_dir(self) -> str:
        return os.path.join(self._dbt_dir, 'contracts')

    @property
    def model_dir(self) -> str:
        return os.path.join(self._dbt_dir, 'models')

    @property
    def codegen_dir(self) -> str:
        return os.path.join(self.model_dir, 'codegen')

    @property
    def dbt_project_yml(self) -> str:
        return os.path.join(self._dbt_dir, 'dbt_project.yml')

    @property
    def new_dbt_project_yml(self) -> str:
        return os.path.join(self._dbt_dir, 'new_dbt_project.yml')

    @functools.cached_property
    def version(self) -> str:
        dbt_project_yml = os.path.join(self._dbt_dir, 'dbt_project.yml')
        with open(dbt_project_yml, 'r') as f:
            project_conf: Dict[str, any] = yaml.safe_load(f)
            dbt_version: str = project_conf['version']

        if dbt_version is None:
            raise ValueError('dbt_project.yml should contains the version field.')

        return dbt_version

    @functools.cached_property
    def project_dirs(self):
        return [f for f in listdir(self.contracts_dir) if os.path.isdir(os.path.join(self.contracts_dir, f))]

    @functools.cached_property
    def contracts_map(self) -> Dict[str, List[Contract]]:
        contracts_map: Dict[str, List[Contract]] = {}
        for project in self.project_dirs:
            contract_paths = glob.glob(os.path.join(self.contracts_dir, project, '*.json'))
            for contract_path in contract_paths:
                with open(contract_path, 'r') as f:
                    if project not in contracts_map:
                        contracts_map[project] = []
                    contracts_map[project].append(cast(Contract, json.loads(f.read())))
        return contracts_map



import logging
import os.path
import pathlib
from typing import Dict, Optional

from bdbt.content import Contract
from bdbt.ethereum.abi.abi_data_type import ABISchema, ABIEventSchema, ABICallSchema


class DbtCodeGenerator:

    def __init__(self, need_udf: bool):
        self.need_udf = need_udf
        self.logger = logging.getLogger(self.__class__.__name__)

    @staticmethod
    def create_file_and_write(filepath: str, content: str):
        if os.path.exists(filepath):
            raise FileExistsError(filepath + ' is exists.')

        with open(filepath, 'w') as f:
            f.write(content)

    def gen_models_for_project(
            self,
            workspace: str,
            project_name: str,
            contract: Contract,
            version: str,
            abi: ABISchema,
    ):
        """
        Generate some dbt model sql files and schema yaml file,
        one model sql file for one event or call.
        """
        project_path = os.path.join(workspace, project_name)
        pathlib.Path(project_path).mkdir(parents=True, exist_ok=True)

        for event in abi.events:
            self.gen_event_dbt_model(project_path, contract, version, event)
        for call in abi.calls:
            self.gen_call_dbt_model(project_path, contract, version, call)

    def gen_udf_for_dbt(
            self,
            dbt_dir: str,
            abi_map: Dict[str, Dict[str, ABISchema]],
            version: str
    ) -> None:
        """
        Generating the external dependency UDF required for the dbt decode contract process, there are some databases
        that are available with a script language runtime, then they do not need to generate additional UDFs via this
        method.
        However, if the database does not support another language in the SQL runtime, then an external UDF may be
        required, such as Spark, which can generate jar packages via this function to enable the use of more complex
        functions in SQL.
        Notes: the best use of this function is to collect all the contract information and call it once to generate all
        the UDFs.

        :param dbt_dir: the absolute path of the dbt project folder
        :param abi_map: project_name -> (contract_name -> abi_schema)
        :param version: the version of the dbt project
        """
        udf_workspace = self.prepare_udf_workspace(dbt_dir)

        # the empty events and empty calls don't need UDF to decode data
        for proj_name, contract_name_to_abi in abi_map.items():
            for contract_name, abi in contract_name_to_abi.items():
                for event in abi.nonempty_events:
                    self.gen_event_udf(udf_workspace, proj_name, contract_name, event)
                for call in abi.nonempty_calls:
                    self.gen_call_udf(udf_workspace, proj_name, contract_name, call)

        self.build_udf(dbt_dir, version)

    def gen_event_dbt_model(
            self,
            project_path: str,
            contract: Contract,
            version: str,
            event: ABIEventSchema
    ):
        raise NotImplementedError()

    def gen_call_dbt_model(
            self,
            project_path: str,
            contract: Contract,
            version: str,
            call: ABICallSchema
    ):
        raise NotImplementedError()

    def gen_event_udf(
            self, udf_workspace: str, project_name, contract_name: str, event: ABIEventSchema
    ) -> None:
        raise NotImplementedError()

    def gen_call_udf(
            self, udf_workspace: str, project_name, contract_name: str, call: ABICallSchema
    ) -> None:
        raise NotImplementedError()

    def prepare_udf_workspace(
            self, dbt_dir: str
    ) -> str:
        raise NotImplementedError()

    def build_udf(
            self, dbt_dir: str, version: str
    ) -> None:
        raise NotImplementedError()

    @staticmethod
    def evt_model_name(
            contract_name: str, event: ABIEventSchema, project_name: Optional[str] = None,
    ) -> str:
        return f'{project_name}_{contract_name}_evt_{event.name}' \
            if project_name is not None else f'{contract_name}_evt_{event.name}'

    @staticmethod
    def call_model_name(
            contract_name: str, call: ABICallSchema, project_name: Optional[str] = None
    ) -> str:
        return f'{project_name}_{contract_name}_call_{call.name}' \
            if project_name is not None else f'{contract_name}_call_{call.name}'



class ABITypeNotValid(Exception):
    """
    We failed to create ABI type
    """

    def __init__(self, message) -> None:
        super().__init__(message)


class TargetItemNotFound(Exception):
    """
    We found no target item in ABI
    """

    def __init__(self, message) -> None:
        super().__init__(message)



from typing import (
    Sequence,
    Union,
    Optional
)

try:
    from typing import TypedDict, Literal
except ImportError:
    from typing_extensions import TypedDict, Literal


# event
class ABIEventElement(TypedDict, total=False):
    indexed: bool
    name: str
    type: str


class ABIEvent(TypedDict, total=False):
    anonymous: bool
    inputs: Sequence[ABIEventElement]
    name: str
    type: Literal["event"]


# function
class ABICallElement(TypedDict, total=False):
    components: Optional[Sequence['ABICallElement']]
    name: str
    type: str


class ABICall(TypedDict, total=False):
    constant: bool
    name: str
    inputs: Sequence[ABICallElement]
    outputs: Sequence[ABICallElement]
    payable: bool
    stateMutability: Literal["pure", "view", "nonpayable", "payable"]
    type: Literal["function", "constructor", "fallback", "receive"]


ABIElement = Union[ABICall, ABIEvent]

# Internal ABI type for safer function parameters
ABI = Sequence[ABIElement]



from bdbt.cli.ethereum_codegen import ethereum_codegen
from bdbt.logging_utils import logging_basic_config

logging_basic_config()

import click


@click.group()
@click.version_option(version='0.2.0')
@click.pass_context
def cli(ctx):
    pass


cli.add_command(ethereum_codegen, "ethereum_codegen")



import json
from typing import cast, List

from bdbt.ethereum.abi.abi_type import ABI, ABIElement


def normalize_abi(abi_json: str) -> ABI:
    """
    Convert a json ABI string to an :class:`ABI` object which uses TypedDict
    """
    abi_json = json.loads(abi_json)
    return cast(ABI, abi_json)


def filter_by_type(type_str: str, contract_abi: ABI) -> List[ABIElement]:
    return [abi for abi in contract_abi if abi['type'] == type_str]


def filter_by_name(name: str, contract_abi: ABI) -> List[ABIElement]:
    return [abi for abi in contract_abi
            if (abi['type'] not in ('fallback', 'constructor', 'receive') and abi['name'] == name)]


def filter_by_type_and_name(name: str, type_str: str, contract_abi: ABI) -> List[ABIElement]:
    return [abi for abi in contract_abi if abi['type'] == type_str and abi['name'] == name]



import os

from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


long_description = read('README.md.md') if os.path.isfile("README.md.md") else ""

setup(
    name='blockchain-dbt',
    version='0.3.9',
    author='Datawaves',
    author_email='team@datawaves.xyz',
    description='The schema transformer and dbt utils for blockchain data.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/datawaves/blockchain-dbt',
    packages=find_packages(exclude=["test.*", "test"]),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ],
    keywords='blockchain, dbt',
    python_requires='>=3.8,<4',
    install_requires=[
        'pyaml==21.10.1',
        'ruamel.yaml==0.17.21',
        'eth-utils==1.10.0',
        'click==7.1.2',
    ],
    extras_require={
        'dev': [
            "pytest~=4.3.0",
            "pyspark==3.2.1"
        ]
    },
    entry_points={
        'console_scripts': [
            'bdbt=bdbt.cli:cli'
        ]
    },
    project_urls={
        'Bug Reports': 'https://github.com/datawaves/blockchain-dbt/issues',
        'Source': 'https://github.com/datawaves/blockchain-dbt',
    },
    tests_require=[
        'nose==1.3.7',
        'unittest2>=1.0.0'
    ]
)



import os.path
import pathlib
from typing import List, Dict

import pyaml

from bdbt.ethereum.abi.abi_data_type import ABISchema
from bdbt.global_type import DbtTable, DbtColumn, DbtModelSchema

evt_base_column = [
    'evt_block_number',
    'evt_block_time',
    'evt_index',
    'evt_tx_hash',
    'contract_address',
    'dt'
]

call_base_column = [
    'call_success',
    'call_block_number',
    'call_block_time',
    'call_trace_address',
    'call_tx_hash',
    'contract_address',
    'dt'
]


class DbtSchemaGenerator:
    @staticmethod
    def gen_dbt_schema(
            workspace: str,
            project_name: str,
            contract_name_to_abi: Dict[str, ABISchema]
    ):
        project_path = os.path.join(workspace, project_name)
        pathlib.Path(project_path).mkdir(parents=True, exist_ok=True)

        models: List[DbtTable] = []
        for contract_name, abi in contract_name_to_abi.items():
            for event in abi.events:
                columns = [DbtColumn(name=i.name) for i in event.inputs]
                columns.extend(DbtColumn(name=i) for i in evt_base_column)
                models.append(DbtTable(name=f'{contract_name}_evt_{event.name}', columns=columns))

            for call in abi.calls:
                columns = [DbtColumn(name=i.name) for i in call.inputs]
                columns.extend([DbtColumn(name=i.name) for i in call.outputs])
                columns.extend(DbtColumn(name=i) for i in call_base_column)
                models.append(DbtTable(name=f'{contract_name}_call_{call.name}', columns=columns))

        schema = DbtModelSchema(models=models)
        schema_path = os.path.join(project_path, 'schema.yml')
        with open(schema_path, 'w') as f:
            # https://docs.getdbt.com/faqs/why-version-2
            f.write('version: 2\n')
            f.write(pyaml.dump(schema, sort_dicts=False))



from typing import TypedDict

from bdbt.ethereum.abi.abi_type import ABI


class Contract(TypedDict, total=False):
    abi: ABI
    address: str
    name: str
    # table / increment
    materialize: str
